@@ -1,12 +1,11 @@
 import binascii
 import datetime
 import math
+import pymongo
 import random
 import time
 import typing
 
-import pymongo
-
 from . import _connector
 
 
@@ -110,7 +109,7 @@ class Lock(
         if not lock_document:
             return None
 
-        if lock_document['expireAt'] < now_date:
+        if lock_document['expireAt'] <= now_date:
             return None
         else:
             expire_time_delta = lock_document['expireAt'] - now_date
