@@ -1,12 +1,12 @@
 import dataclasses
 import datetime
-import logging
+import orjson
 import socket
 import sys
 import traceback
 import typing
 
-import orjson
+import logging
 
 
 class LogstashHandler(
@@ -63,7 +63,7 @@ class LogstashHandler(
             'emitter': {
                 'hostname': self.hostname,
                 'ipaddress': self.ipaddress,
-                'filename': record.filename,
+                'pathname': record.pathname,
                 'function': record.funcName,
                 'line': record.lineno,
             },
@@ -105,5 +105,5 @@ class LogstashHandler(
                 socket_connection.settimeout(self.timeout)
                 socket_connection.connect(self.address)
                 socket_connection.sendall(encoded_message)
-        except Exception:
-            traceback.print_exc()
+        except Exception as exception:
+            print(f'sending log entry to the logstash server has failed: {exception}')
