@@ -1,15 +1,15 @@
 import argparse
-import logging
 import multiprocessing
 import multiprocessing.context
-import os
+import psutil
 import shlex
+import signal
 import subprocess
 import sys
 import time
 import typing
 
-import psutil
+import logging
 
 
 class SupervisedWorker:
@@ -99,6 +99,10 @@ class Supervisor:
         self.concurrent_workers = concurrent_workers
         self.max_worker_memory_usage = max_worker_memory_usage
 
+        self.stop_process_has_started = False
+
+        self.supevisor_process = psutil.Process()
+
         self.extra_signature: typing.Dict[str, typing.Any] = {
             'supervisor': {
                 'worker_module_name': self.worker_module_name,
@@ -154,6 +158,8 @@ class Supervisor:
                         worker=worker,
                     )
 
+                self.clean_zombies()
+
                 time.sleep(0.5)
 
             self.logger.info(
@@ -187,72 +193,69 @@ class Supervisor:
     ) -> None:
         worker_has_finished_running = worker.process.poll() is not None
         if worker_has_finished_running:
+            extra_signature = self.extra_signature.copy()
+
             try:
                 worker_summary = worker.get_summary()
+                extra_signature['summary'] = worker_summary
             except Exception as exception:
                 self.logger.error(
                     msg=f'could not receive supervised_worker\'s summary: {exception}',
                     extra=self.extra_signature,
                 )
                 worker_summary = None
+                extra_signature['summary'] = {}
 
             try:
-                os.waitpid(worker.process.pid, 0)
-            except ChildProcessError:
+                worker.process.wait()
+            except (
+                ChildProcessError,
+                OSError,
+            ):
                 pass
 
-            if worker.process.returncode == 0:
-                extra_signature = self.extra_signature.copy()
-                extra_signature['summary'] = worker_summary if worker_summary else {}
-                if worker_summary and (
-                    worker_summary['executor_exception'] or
-                    worker_summary['initialize_exception'] or
-                    worker_summary['finalize_exception']
-                ):
-                    self.logger.critical(
-                        msg='worker internal execution has failed',
-                        extra=extra_signature,
+            worker_return_code = worker.process.returncode
+            if worker_summary and 'return_code' in worker_summary:
+                worker_return_code = worker_summary['return_code']
+                if 'unkillable_threads' in worker_summary:
+                    unkillable_threads = worker_summary['unkillable_threads']
+                    self.logger.warning(
+                        msg=f'worker({worker.process.pid}) had unkillable_threads: {unkillable_threads}',
+                        extra=self.extra_signature,
                     )
-                else:
-                    self.logger.info(
-                        msg='worker has finished successfully',
-                        extra=extra_signature,
-                    )
-            elif worker.process.returncode == 1:
-                extra_signature = self.extra_signature.copy()
-                extra_signature['exception'] = worker_summary if worker_summary else {}
-                exception_message = extra_signature['exception'].get('message', None)
 
-                self.logger.critical(
-                    msg=f'worker execution has failed with the following exception: {exception_message}',
+            if worker_return_code == 0:
+                self.logger.info(
+                    msg=f'worker({worker.process.pid}) has finished successfully',
                     extra=extra_signature,
                 )
-            elif worker.process.returncode == 2:
+            elif worker_return_code == 1:
+                self.logger.critical(
+                    msg=f'worker({worker.process.pid}) execution has failed with the following exception: {extra_signature["summary"].get("exception")}',
+                    extra=extra_signature,
+                )
+            elif worker_return_code == 2:
                 self.logger.critical(
                     msg=f'could not load worker module: {self.worker_module_name}',
                     extra=self.extra_signature,
                 )
 
                 sys.exit(1)
-            elif worker.process.returncode == 3:
+            elif worker_return_code == 3:
                 self.logger.critical(
                     msg=f'could not find worker class: {self.worker_module_name}.{self.worker_class_name}',
                     extra=self.extra_signature,
                 )
 
                 sys.exit(1)
-            elif worker.process.returncode == 4:
-                extra_signature = self.extra_signature.copy()
-                extra_signature['summary'] = worker_summary if worker_summary else {}
+            elif worker_return_code == 4:
                 self.logger.info(
-                    msg='worker has requested to respawn',
+                    msg=f'worker({worker.process.pid}) has requested to respawn',
                     extra=extra_signature,
                 )
-            elif worker.process.returncode == 5:
-                extra_signature = self.extra_signature.copy()
-                extra_signature['summary'] = worker_summary if worker_summary else {}
+            elif worker_return_code == 5:
                 self.logger.info(
-                    msg='worker has requested to stop',
+                    msg=f'worker({worker.process.pid}) has requested to stop',
                     extra=extra_signature,
                 )
 
@@ -260,13 +263,18 @@ class Supervisor:
                     worker=worker,
                 )
 
+                self.stop_process_has_started = True
+
                 return
-            else:
-                extra_signature = self.extra_signature.copy()
-                extra_signature['exception'] = worker_summary if worker_summary else {}
-                extra_signature['return_code'] = worker.process.returncode
+            elif worker_return_code == 6:
                 self.logger.critical(
-                    msg=f'worker execution has been interrupted with return code: {worker.process.returncode}',
+                    msg=f'worker({worker.process.pid}) internal execution has failed',
+                    extra=extra_signature,
+                )
+            else:
+                extra_signature['return_code'] = worker_return_code
+                self.logger.critical(
+                    msg=f'worker({worker.process.pid}) execution has been interrupted with return code: {worker_return_code}',
                     extra=self.extra_signature,
                 )
 
@@ -276,40 +284,75 @@ class Supervisor:
         else:
             if self.max_worker_memory_usage and worker.get_rss_memory() > self.max_worker_memory_usage:
                 self.logger.warning(
-                    msg=f'worker exceeded the maximum memory limit: pid: {worker.process.pid}, rss: {worker.get_rss_memory()}',
+                    msg=f'worker({worker.process.pid}) exceeded the maximum memory limit: {worker.get_rss_memory()}',
                     extra=self.extra_signature,
                 )
                 self.respawn_a_worker(
                     worker=worker,
                 )
 
+        if self.stop_process_has_started:
+            try:
+                worker.psutil_obj.send_signal(
+                    sig=signal.SIGUSR1,
+                )
+            except psutil.NoSuchProcess:
+                pass
+
     def respawn_a_worker(
         self,
         worker: SupervisedWorker,
     ) -> None:
         worker.kill()
+
         self.current_workers.remove(worker)
-        new_worker = SupervisedWorker(
-            worker_module_name=self.worker_module_name,
-            worker_class_name=self.worker_class_name,
-        )
-        self.logger.info(
-            msg=f'spawned a new worker at pid: {new_worker.process.pid}',
-            extra=self.extra_signature,
-        )
-        self.current_workers.append(new_worker)
+
+        if self.stop_process_has_started:
+            self.logger.info(
+                msg='not going to respawn since the stop process has started',
+                extra=self.extra_signature,
+            )
+        else:
+            new_worker = SupervisedWorker(
+                worker_module_name=self.worker_module_name,
+                worker_class_name=self.worker_class_name,
+            )
+            self.logger.info(
+                msg=f'worker({worker.process.pid}) was respawned as worker({new_worker.process.pid})',
+                extra=self.extra_signature,
+            )
+            self.current_workers.append(new_worker)
 
     def stop_a_worker(
         self,
         worker: SupervisedWorker,
     ) -> None:
         worker.kill()
+
         self.current_workers.remove(worker)
         self.logger.info(
-            msg=f'worker has stopped: {worker.process.pid}',
+            msg=f'worker({worker.process.pid}) has stopped',
             extra=self.extra_signature,
         )
 
+    def clean_zombies(
+        self,
+    ) -> None:
+        supervisor_zombie_children = [
+            child_process
+            for child_process in self.supevisor_process.children()
+            if child_process.status() == psutil.STATUS_ZOMBIE
+        ]
+        if supervisor_zombie_children:
+            self.logger.info(
+                msg=f'cleaning {len(supervisor_zombie_children)} zombies',
+                extra=self.extra_signature,
+            )
+            psutil.wait_procs(
+                procs=supervisor_zombie_children,
+                timeout=1.0,
+            )
+
 
 def main() -> None:
     parser = argparse.ArgumentParser(
